@@ -17,12 +17,12 @@
 package org.apache.karaf.features.internal.region;
 
 import java.io.IOException;
-import java.io.InputStream;
 import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Collections;
 import java.util.HashMap;
 import java.util.HashSet;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
@@ -32,8 +32,15 @@ import java.util.jar.Manifest;
 import java.util.zip.ZipEntry;
 import java.util.zip.ZipInputStream;
 
+import org.apache.felix.utils.collections.StringArrayMap;
 import org.apache.felix.utils.manifest.Clause;
 import org.apache.felix.utils.manifest.Parser;
+import org.apache.felix.utils.repository.BaseRepository;
+import org.apache.felix.utils.resource.CapabilityImpl;
+import org.apache.felix.utils.resource.RequirementImpl;
+import org.apache.felix.utils.resource.ResourceBuilder;
+import org.apache.felix.utils.resource.ResourceImpl;
+import org.apache.felix.utils.resource.SimpleFilter;
 import org.apache.felix.utils.version.VersionRange;
 import org.apache.felix.utils.version.VersionTable;
 import org.apache.karaf.features.BundleInfo;
@@ -44,19 +51,12 @@ import org.apache.karaf.features.FeaturesNamespaces;
 import org.apache.karaf.features.FeaturesService;
 import org.apache.karaf.features.Library;
 import org.apache.karaf.features.ScopeFilter;
-import org.apache.karaf.features.internal.download.DownloadCallback;
 import org.apache.karaf.features.internal.download.DownloadManager;
 import org.apache.karaf.features.internal.download.Downloader;
 import org.apache.karaf.features.internal.download.StreamProvider;
-import org.apache.karaf.features.internal.repository.BaseRepository;
-import org.apache.karaf.features.internal.resolver.CapabilityImpl;
 import org.apache.karaf.features.internal.resolver.FeatureResource;
-import org.apache.karaf.features.internal.resolver.RequirementImpl;
 import org.apache.karaf.features.internal.resolver.ResolverUtil;
-import org.apache.karaf.features.internal.resolver.ResourceBuilder;
-import org.apache.karaf.features.internal.resolver.ResourceImpl;
 import org.apache.karaf.features.internal.resolver.ResourceUtils;
-import org.apache.karaf.features.internal.resolver.SimpleFilter;
 import org.apache.karaf.features.internal.service.Overrides;
 import org.osgi.framework.BundleException;
 import org.osgi.framework.Version;
@@ -77,6 +77,9 @@ import static org.osgi.framework.namespace.IdentityNamespace.CAPABILITY_VERSION_
 import static org.osgi.framework.namespace.IdentityNamespace.IDENTITY_NAMESPACE;
 import static org.osgi.resource.Namespace.REQUIREMENT_FILTER_DIRECTIVE;
 
+/**
+ * A {@link Resource} representing ...
+ */
 public class Subsystem extends ResourceImpl {
 
     private static final String ALL_FILTER = "(|(!(all=*))(all=*))";
@@ -99,20 +102,49 @@ public class Subsystem extends ResourceImpl {
                     IDENTITY_NAMESPACE,
                     Collections.singleton(SUBSYSTEM_FILTER));
 
+    // name of the subsystem: region or region#feature[-version]
     private final String name;
+    // works only with feature scoping. region subsystems by default accept deps
     private final boolean acceptDependencies;
+    // parent Subsystem for child subsystems representing child regions or regions' features
     private final Subsystem parent;
+    // feature for Subsystem representing a feature
     private final Feature feature;
+
     private final boolean mandatory;
+
     private final List<Subsystem> children = new ArrayList<>();
+
+    // a set of filters applied when child subsystem needs capabilities from parent subsystem
     private final Map<String, Set<String>> importPolicy;
+    // a set of filters applied when parent subsystem needs capabilities from child subsystem
     private final Map<String, Set<String>> exportPolicy;
+
+    // contains subsystems representing features of this region, child subsystems for child regions, system resources(?),
+    // bundle resources added explicitly as reqs for this Subsystem, feature resources for subsystems representing
+    // features, ...
     private final List<Resource> installable = new ArrayList<>();
+
+    // mapping from "symbolic-name|version" to a DependencyInfo wrapping a Resource
+    // <bundle dependency="false"> are collected directly in feature's subsystem
+    // <bundle dependency="true"> are collected in first parent subsystem of feature or in subsystem of scoped feature
     private final Map<String, DependencyInfo> dependencies = new HashMap<>();
+    // non-mandatory dependant features (<feature>/<feature>) collected from current and child subsystems representing
+    // features (unless some subsystem for feature has <scoping acceptDependencies="true">)
     private final List<Requirement> dependentFeatures = new ArrayList<>();
 
+    // direct bundle URI dependencies - not added by FeaturesService, but used in startup stage of assembly builder
+    // these bundles will be downloaded
     private final List<String> bundles = new ArrayList<>();
 
+    /**
+     * Constructs root subsystem {@link Resource} for {@link FeaturesService#ROOT_REGION} that imports/exports only
+     * caps/reqs with <code>(type=karaf.subsystem)</code>.
+     * Root subsystem by default accepts dependencies - will gather dependant features of child feature subsystems,
+     * effectively _flattening_ the set of features within single region's subsystem.
+     *
+     * @param name The name of the subsystem.
+     */
     public Subsystem(String name) {
         super(name, TYPE_SUBSYSTEM, Version.emptyVersion);
         this.name = name;
@@ -124,6 +156,17 @@ public class Subsystem extends ResourceImpl {
         this.mandatory = true;
     }
 
+    /**
+     * Constructs subsystem for a feature that either imports/exports all caps or (see {@link Feature#getScoping()})
+     * has configurable import/export policy + <code>(|(type=karaf.subsystem)(type=karaf.feature))</code> filter in
+     * {@link org.osgi.framework.namespace.IdentityNamespace#IDENTITY_NAMESPACE}.
+     * Such subsystem requires <code>type=karaf.feature; osgi.identity=feature-name[; version=feature-version]</code>.
+     *
+     * @param name The subsystem name.
+     * @param feature The feature.
+     * @param parent The parent subsystem.
+     * @param mandatory True if mandatory, false else.
+     */
     public Subsystem(String name, Feature feature, Subsystem parent, boolean mandatory) {
         super(name, TYPE_SUBSYSTEM, Version.emptyVersion);
         this.name = name;
@@ -147,6 +190,15 @@ public class Subsystem extends ResourceImpl {
                 new VersionRange(VersionTable.getVersion(feature.getVersion()), true));
     }
 
+    /**
+     * Constructs child subsystem {@link Resource} for {@link FeaturesService#ROOT_REGION}'s child
+     * that imports all caps and exports only caps with <code>(type=karaf.subsystem)</code>.
+     *
+     * @param name The subsystem name.
+     * @param parent The parent subsystem.
+     * @param acceptDependencies True to accept dependencies, false else.
+     * @param mandatory True to mandatory, false else.
+     */
     public Subsystem(String name, Subsystem parent, boolean acceptDependencies, boolean mandatory) {
         super(name, TYPE_SUBSYSTEM, Version.emptyVersion);
         this.name = name;
@@ -199,6 +251,12 @@ public class Subsystem extends ResourceImpl {
         return feature;
     }
 
+    /**
+     * Create child subsystem for this subsystem. Child will become parent's mandatory requirement to force its resolution.
+     *
+     * @param name The subsystem name.
+     * @param acceptDependencies True to accept dependencies, false else.
+     */
     public Subsystem createSubsystem(String name, boolean acceptDependencies) {
         if (feature != null) {
             throw new UnsupportedOperationException("Can not create application subsystems inside a feature subsystem");
@@ -270,29 +328,38 @@ public class Subsystem extends ResourceImpl {
     }
 
     @SuppressWarnings("InfiniteLoopStatement")
-    public void build(Collection<Feature> features) throws Exception {
-        doBuild(features, true);
+    public void build(Map<String, List<Feature>> allFeatures) throws Exception {
+        doBuild(allFeatures, true);
     }
 
-    private void doBuild(Collection<Feature> features, boolean mandatory) throws Exception {
+    private void doBuild(Map<String, List<Feature>> allFeatures, boolean mandatory) throws Exception {
         for (Subsystem child : children) {
-            child.doBuild(features, true);
+            child.doBuild(allFeatures, true);
         }
         if (feature != null) {
+            // each dependant feature becomes a non-mandatory (why?) requirement of first parent that
+            // accepts dependencies
             for (Dependency dep : feature.getDependencies()) {
+                if (dep.isBlacklisted()) {
+                    continue;
+                }
                 Subsystem ss = this;
                 while (!ss.isAcceptDependencies()) {
                     ss = ss.getParent();
                 }
                 ss.requireFeature(dep.getName(), dep.getVersion(), false);
             }
+            // each conditional feature becomes a child subsystem of this feature's subsystem
             for (Conditional cond : feature.getConditional()) {
+                if (cond.isBlacklisted()) {
+                    continue;
+                }
                 Feature fcond = cond.asFeature();
                 String ssName = this.name + "#" + (fcond.hasVersion() ? fcond.getName() + "-" + fcond.getVersion() : fcond.getName());
                 Subsystem fs = getChild(ssName);
                 if (fs == null) {
                     fs = new Subsystem(ssName, fcond, this, true);
-                    fs.doBuild(features, false);
+                    fs.doBuild(allFeatures, false);
                     installable.add(fs);
                     children.add(fs);
                 }
@@ -306,20 +373,21 @@ public class Subsystem extends ResourceImpl {
             if (requirements.isEmpty()) {
                 break;
             }
+            // for each feature requirement on this subsystem (osgi.identity;type=karaf.feature), we create a
+            // Subsystem representing mandatory feature.
             for (Requirement requirement : requirements) {
                 String name = (String) requirement.getAttributes().get(IDENTITY_NAMESPACE);
                 String type = (String) requirement.getAttributes().get(CAPABILITY_TYPE_ATTRIBUTE);
                 VersionRange range = (VersionRange) requirement.getAttributes().get(CAPABILITY_VERSION_ATTRIBUTE);
-                if (TYPE_FEATURE.equals(type)) {
-                    for (Feature feature : features) {
-                        if (feature.getName().equals(name)
-                                && (range == null || range.contains(VersionTable.getVersion(feature.getVersion())))) {
-                            if (feature != this.feature) {
+                if (TYPE_FEATURE.equals(type) && allFeatures.containsKey(name)) {
+                    for (Feature feature : allFeatures.get(name)) {
+                        if (range == null || range.contains(VersionTable.getVersion(feature.getVersion()))) {
+                            if (feature != this.feature && !feature.isBlacklisted()) {
                                 String ssName = this.name + "#" + (feature.hasVersion() ? feature.getName() + "-" + feature.getVersion() : feature.getName());
                                 Subsystem fs = getChild(ssName);
                                 if (fs == null) {
                                     fs = new Subsystem(ssName, feature, this, mandatory && !SubsystemResolveContext.isOptional(requirement));
-                                    fs.build(features);
+                                    fs.build(allFeatures);
                                     installable.add(fs);
                                     children.add(fs);
                                 }
@@ -348,10 +416,12 @@ public class Subsystem extends ResourceImpl {
                 String[] p = prereq.split("/");
                 if (feature.getName().equals(p[0])
                         && VersionRange.parseVersionRange(p[1]).contains(Version.parseVersion(feature.getVersion()))) {
+                    // our feature is already among prerequisites, so ...
                     match = true;
                     break;
                 }
             }
+            // ... we won't be adding its prerequisites - they'll be handled after another PartialDeploymentException
             if (!match) {
                 for (Dependency dep : feature.getDependencies()) {
                     if (dep.isPrerequisite()) {
@@ -362,95 +432,127 @@ public class Subsystem extends ResourceImpl {
         }
     }
 
+    /**
+     * Downloads bundles for all the features in current and child subsystems. But also collects bundles
+     * as {@link DependencyInfo}.
+     *
+     * @param manager The {@link DownloadManager} to use.
+     * @param featureResolutionRange The feature resolution range to use.
+     * @param serviceRequirements The {@link FeaturesService.ServiceRequirementsBehavior} behavior to use.
+     * @param repos The {@link RepositoryManager} to use.
+     * @param callback The {@link SubsystemResolverCallback} to use.
+     */
     @SuppressWarnings("InfiniteLoopStatement")
     public void downloadBundles(DownloadManager manager,
-                                Set<String> overrides,
                                 String featureResolutionRange,
-                                final String serviceRequirements,
-                                RepositoryManager repos) throws Exception {
+                                final FeaturesService.ServiceRequirementsBehavior serviceRequirements,
+                                RepositoryManager repos,
+                                SubsystemResolverCallback callback) throws Exception {
         for (Subsystem child : children) {
-            child.downloadBundles(manager, overrides, featureResolutionRange, serviceRequirements, repos);
+            child.downloadBundles(manager, featureResolutionRange, serviceRequirements, repos, callback);
         }
-        final Map<String, ResourceImpl> bundles = new ConcurrentHashMap<>();
-        final Downloader downloader = manager.createDownloader();
+
+        // collect BundleInfos for given feature - both direct <feature>/<bundle>s and <feature>/<conditional>/<bundle>s
         final Map<BundleInfo, Conditional> infos = new HashMap<>();
+        final Downloader downloader = manager.createDownloader();
         if (feature != null) {
             for (Conditional cond : feature.getConditional()) {
-                for (final BundleInfo bi : cond.getBundles()) {
-                    infos.put(bi, cond);
+                if (!cond.isBlacklisted()) {
+                    for (final BundleInfo bi : cond.getBundles()) {
+                        // bundles from conditional features will be added as non-mandatory requirements
+                        infos.put(bi, cond);
+                    }
                 }
             }
             for (BundleInfo bi : feature.getBundles()) {
                 infos.put(bi, null);
             }
         }
-        boolean removeServiceRequirements;
-        if (FeaturesService.SERVICE_REQUIREMENTS_DISABLE.equals(serviceRequirements)) {
-            removeServiceRequirements = true;
-        } else if (feature != null && FeaturesService.SERVICE_REQUIREMENTS_DEFAULT.equals(serviceRequirements)) {
-            removeServiceRequirements = !FeaturesNamespaces.URI_1_3_0.equals(feature.getNamespace())
-                                     && !FeaturesNamespaces.URI_1_4_0.equals(feature.getNamespace());
-        } else {
-            removeServiceRequirements = false;
+
+        // features model doesn't have blacklisted entries removed, but marked as blacklisted - we now don't have
+        // to download them
+        //infos.keySet().removeIf(Blacklisting::isBlacklisted);
+        for (Iterator<BundleInfo> iterator = infos.keySet().iterator(); iterator.hasNext(); ) {
+            BundleInfo bi = iterator.next();
+            if (bi.isBlacklisted()) {
+                iterator.remove();
+                if (callback != null) {
+                    callback.bundleBlacklisted(bi);
+                }
+            }
         }
+
+        // all downloaded bundles
+        final Map<String, ResourceImpl> bundles = new ConcurrentHashMap<>();
+        // resources for locations that were overriden in OSGi mode - to check whether the override should actually
+        // take place, by checking resource's headers
+        final Map<String, ResourceImpl> overrides = new ConcurrentHashMap<>();
+
+        boolean removeServiceRequirements = serviceRequirementsBehavior(feature, serviceRequirements);
+
+        // download collected BundleInfo locations
         for (Map.Entry<BundleInfo, Conditional> entry : infos.entrySet()) {
             final BundleInfo bi = entry.getKey();
             final String loc = bi.getLocation();
-            downloader.download(loc, new DownloadCallback() {
-                @Override
-                public void downloaded(StreamProvider provider) throws Exception {
-                    ResourceImpl res = createResource(loc, getMetadata(provider), removeServiceRequirements);
-                    bundles.put(loc, res);
+            downloader.download(loc, provider -> {
+                // always download location (could be overriden)
+                ResourceImpl resource = createResource(loc, getMetadata(provider), removeServiceRequirements);
+                bundles.put(loc, resource);
+
+                if (bi.isOverriden() == BundleInfo.BundleOverrideMode.OSGI) {
+                    // also download original from original bundle URI to check if we should override by comparing
+                    // symbolic name - requires MANIFEST.MF header access. If there should be no override, we'll get
+                    // back to original URI
+                    downloader.download(bi.getOriginalLocation(), provider2 -> {
+                        ResourceImpl originalResource = createResource(bi.getOriginalLocation(),
+                                getMetadata(provider2), removeServiceRequirements);
+                        bundles.put(bi.getOriginalLocation(), originalResource);
+                        // an entry in overrides map means that given location was overriden
+                        overrides.put(loc, originalResource);
+                    });
                 }
             });
         }
+        // download direct bundle: requirements - without consulting overrides
         for (Clause bundle : Parser.parseClauses(this.bundles.toArray(new String[this.bundles.size()]))) {
             final String loc = bundle.getName();
-            downloader.download(loc, new DownloadCallback() {
-                @Override
-                public void downloaded(StreamProvider provider) throws Exception {
-                    ResourceImpl res = createResource(loc, getMetadata(provider), removeServiceRequirements);
-                    bundles.put(loc, res);
-                }
-            });
-        }
-        for (String override : overrides) {
-            final String loc = Overrides.extractUrl(override);
-            downloader.download(loc, new DownloadCallback() {
-                @Override
-                public void downloaded(StreamProvider provider) throws Exception {
-                    ResourceImpl res = createResource(loc, getMetadata(provider), removeServiceRequirements);
-                    bundles.put(loc, res);
-                }
-            });
+            downloader.download(loc, provider -> bundles.put(loc, createResource(loc, getMetadata(provider), removeServiceRequirements)));
         }
+        // we *don't* have to download overrides separately - they're already taken into account from processed model
+
+        // download additional libraries - only exported, so they're capabilities are taken into account during
+        // resolution process
         if (feature != null) {
             for (Library library : feature.getLibraries()) {
                 if (library.isExport()) {
                     final String loc = library.getLocation();
-                    downloader.download(loc, new DownloadCallback() {
-                        @Override
-                        public void downloaded(StreamProvider provider) throws Exception {
-                            ResourceImpl res = createResource(loc, getMetadata(provider), removeServiceRequirements);
-                            bundles.put(loc, res);
-                        }
-                    });
+                    downloader.download(loc, provider -> bundles.put(loc, createResource(loc, getMetadata(provider), removeServiceRequirements)));
                 }
             }
         }
         downloader.await();
+
+        // opposite to what we had before. Currently bundles are already overriden at model level, but
+        // as we finally have access to headers, we can compare symbolic names and if override mode is OSGi, then
+        // we can restore original resource if there should be no override.
         Overrides.override(bundles, overrides);
+
         if (feature != null) {
             // Add conditionals
             Map<Conditional, Resource> resConds = new HashMap<>();
             for (Conditional cond : feature.getConditional()) {
+                if (cond.isBlacklisted()) {
+                    continue;
+                }
                 FeatureResource resCond = FeatureResource.build(feature, cond, featureResolutionRange, bundles);
+                // feature's subsystem will optionally require conditional feature resource
                 addIdentityRequirement(this, resCond, false);
+                // but it's a mandatory requirement in other way
                 addIdentityRequirement(resCond, this, true);
                 installable.add(resCond);
                 resConds.put(cond, resCond);
             }
-            // Add features
+            // Add features and make it require given subsystem that represents logical feature requirement
             FeatureResource resFeature = FeatureResource.build(feature, featureResolutionRange, bundles);
             addIdentityRequirement(resFeature, this);
             installable.add(resFeature);
@@ -461,27 +563,29 @@ public class Subsystem extends ResourceImpl {
                 final Conditional cond = entry.getValue();
                 ResourceImpl res = bundles.get(loc);
                 int sl = bi.getStartLevel() <= 0 ? feature.getStartLevel() : bi.getStartLevel();
-                if (bi.isDependency()) {
-                    addDependency(res, false, bi.isStart(), sl);
-                } else {
-                    doAddDependency(res, cond == null, bi.isStart(), sl);
-                }
                 if (cond != null) {
+                    // bundle of conditional feature will have mandatory requirement on it
                     addIdentityRequirement(res, resConds.get(cond), true);
                 }
+                boolean mandatory = !bi.isDependency() && cond == null;
+                if (bi.isDependency()) {
+                    addDependency(res, mandatory, bi.isStart(), sl, bi.isBlacklisted());
+                } else {
+                    doAddDependency(res, mandatory, bi.isStart(), sl, bi.isBlacklisted());
+                }
             }
             for (Library library : feature.getLibraries()) {
                 if (library.isExport()) {
                     final String loc = library.getLocation();
                     ResourceImpl res = bundles.get(loc);
-                    addDependency(res, false, false, 0);
+                    addDependency(res, false, false, 0, false);
                 }
             }
             for (String uri : feature.getResourceRepositories()) {
                 BaseRepository repo = repos.getRepository(feature.getRepositoryUrl(), uri);
                 for (Resource resource : repo.getResources()) {
                     ResourceImpl res = cloneResource(resource);
-                    addDependency(res, false, true, 0);
+                    addDependency(res, false, true, 0, false);
                 }
             }
         }
@@ -489,6 +593,7 @@ public class Subsystem extends ResourceImpl {
             final String loc = bundle.getName();
             boolean dependency = Boolean.parseBoolean(bundle.getAttribute("dependency"));
             boolean start = bundle.getAttribute("start") == null || Boolean.parseBoolean(bundle.getAttribute("start"));
+            boolean blacklisted = bundle.getAttribute("blacklisted") != null && Boolean.parseBoolean(bundle.getAttribute("blacklisted"));
             int startLevel = 0;
             try {
                 startLevel = Integer.parseInt(bundle.getAttribute("start-level"));
@@ -496,24 +601,44 @@ public class Subsystem extends ResourceImpl {
                 // Ignore
             }
             if (dependency) {
-                addDependency(bundles.get(loc), false, start, startLevel);
+                addDependency(bundles.get(loc), false, start, startLevel, blacklisted);
             } else {
-                doAddDependency(bundles.get(loc), true, start, startLevel);
+                doAddDependency(bundles.get(loc), true, start, startLevel, blacklisted);
+                // non dependency bundle will be added as osgi.identity req on type=osgi.bundle
                 addIdentityRequirement(this, bundles.get(loc));
             }
         }
         // Compute dependencies
         for (DependencyInfo info : dependencies.values()) {
             installable.add(info.resource);
+            // bundle resource will have a requirement on its feature's subsystem too
+            // when bundle is declared with dependency="true", it will have a requirement on its region's subsystem
             addIdentityRequirement(info.resource, this, info.mandatory);
         }
     }
 
+    /**
+     * How to handle requirements from {@link org.osgi.namespace.service.ServiceNamespace#SERVICE_NAMESPACE} for
+     * given feature.
+     */
+    private boolean serviceRequirementsBehavior(Feature feature, FeaturesService.ServiceRequirementsBehavior serviceRequirements) {
+        if (FeaturesService.ServiceRequirementsBehavior.Disable == serviceRequirements) {
+            return true;
+        } else if (feature != null && FeaturesService.ServiceRequirementsBehavior.Default == serviceRequirements) {
+            return FeaturesNamespaces.URI_1_0_0.equals(feature.getNamespace())
+                    || FeaturesNamespaces.URI_1_1_0.equals(feature.getNamespace())
+                    || FeaturesNamespaces.URI_1_2_0.equals(feature.getNamespace())
+                    || FeaturesNamespaces.URI_1_2_1.equals(feature.getNamespace());
+        } else {
+            return false;
+        }
+    }
+
     ResourceImpl cloneResource(Resource resource) {
         ResourceImpl res = new ResourceImpl();
         for (Capability cap : resource.getCapabilities(null)) {
             res.addCapability(new CapabilityImpl(res, cap.getNamespace(),
-                    new HashMap<>(cap.getDirectives()), new HashMap<>(cap.getAttributes())));
+                    new StringArrayMap<>(cap.getDirectives()), new StringArrayMap<>(cap.getAttributes())));
         }
         for (Requirement req : resource.getRequirements(null)) {
             SimpleFilter sf;
@@ -525,7 +650,7 @@ public class Subsystem extends ResourceImpl {
                 sf = SimpleFilter.convert(req.getAttributes());
             }
             res.addRequirement(new RequirementImpl(res, req.getNamespace(),
-                    new HashMap<>(req.getDirectives()), new HashMap<>(req.getAttributes()), sf));
+                    new StringArrayMap<>(req.getDirectives()), new StringArrayMap<>(req.getAttributes()), sf));
         }
         return res;
     }
@@ -538,7 +663,7 @@ public class Subsystem extends ResourceImpl {
             while ((entry = zis.getNextEntry()) != null) {
                 if (MANIFEST_NAME.equals(entry.getName())) {
                     Attributes attributes = new Manifest(zis).getMainAttributes();
-                    Map<java.lang.String, java.lang.String> headers = new HashMap<java.lang.String, java.lang.String>();
+                    Map<java.lang.String, java.lang.String> headers = new HashMap<>();
                     for (Map.Entry attr : attributes.entrySet()) {
                         headers.put(attr.getKey().toString(), attr.getValue().toString());
                     }
@@ -549,32 +674,88 @@ public class Subsystem extends ResourceImpl {
         throw new IllegalArgumentException("Resource " + provider.getUrl() + " does not contain a manifest");
     }
 
-    void addDependency(ResourceImpl resource, boolean mandatory, boolean start, int startLevel) {
+    /**
+     * Adds a {@link Resource} as dependency if this subsystem {@link Subsystem#isAcceptDependencies() accepts dependencies},
+     * otherwise, the dependency is added to parent subsystem, effectively searching for first parent subsystem representing
+     * region or scoped feature.
+     */
+    void addDependency(ResourceImpl resource, boolean mandatory, boolean start, int startLevel, boolean blacklisted) {
         if (isAcceptDependencies()) {
-            doAddDependency(resource, mandatory, start, startLevel);
+            doAddDependency(resource, mandatory, start, startLevel, blacklisted);
         } else {
-            parent.addDependency(resource, mandatory, start, startLevel);
+            parent.addDependency(resource, mandatory, start, startLevel, blacklisted);
         }
     }
 
-    private void doAddDependency(ResourceImpl resource, boolean mandatory, boolean start, int startLevel) {
+    /**
+     * Adds a {@link Resource} to this subsystem.
+     */
+    private void doAddDependency(ResourceImpl resource, boolean mandatory, boolean start, int startLevel, boolean blacklisted) {
         String id = ResolverUtil.getSymbolicName(resource) + "|" + ResolverUtil.getVersion(resource);
-        DependencyInfo info = dependencies.computeIfAbsent(id, k -> new DependencyInfo());
-        info.resource = resource;
-        info.mandatory |= mandatory;
-        info.start |= start;
-        if (info.startLevel > 0 && startLevel > 0) {
-            info.startLevel = Math.min(info.startLevel, startLevel);
-        } else {
-            info.startLevel = Math.max(info.startLevel, startLevel);
-        }
+        DependencyInfo info = new DependencyInfo(resource, mandatory, start, startLevel, blacklisted);
+        dependencies.merge(id, info, this::merge);
     }
 
+    /**
+     * Merges two dependencies by taking lower start level, stronger <code>mandatory</code> option and stronger
+     * <code>start</code> option.
+     */
+    private DependencyInfo merge(DependencyInfo di1, DependencyInfo di2) {
+        DependencyInfo info = new DependencyInfo();
+        if (di1.resource != di2.resource) {
+            Requirement r1 = getFirstIdentityReq(di1.resource);
+            Requirement r2 = getFirstIdentityReq(di2.resource);
+            if (r1 == null) {
+                info.resource = di1.resource;
+            } else if (r2 == null) {
+                info.resource = di2.resource;
+            } else {
+                String id = ResolverUtil.getSymbolicName(di1.resource) + "/" + ResolverUtil.getVersion(di1.resource);
+                throw new IllegalStateException("Resource " + id + " is duplicated on subsystem " + this.toString() + ". First resource requires " + r1 + " while the second requires " + r2);
+            }
+        } else {
+            info.resource = di1.resource;
+        }
+        info.mandatory = di1.mandatory | di2.mandatory;
+        info.start = di1.start | di2.start;
+        if (di1.startLevel > 0 && di2.startLevel > 0) {
+            info.startLevel = Math.min(di1.startLevel, di2.startLevel);
+        } else {
+            info.startLevel = Math.max(di1.startLevel, di2.startLevel);
+        }
+        return info;
+    }
+
+    private RequirementImpl getFirstIdentityReq(ResourceImpl resource) {
+        for (Requirement r : resource.getRequirements(null)) {
+            if (IDENTITY_NAMESPACE.equals(r.getNamespace())) {
+                return (RequirementImpl) r;
+            }
+        }
+        return null;
+    }
+
+    /**
+     * TODO DOCUMENT: More generic than just {@link BundleInfo}
+     */
     class DependencyInfo implements BundleInfo {
         ResourceImpl resource;
         boolean mandatory;
         boolean start;
         int startLevel;
+        boolean blacklisted;
+        BundleInfo.BundleOverrideMode overriden;
+
+        public DependencyInfo() {
+        }
+
+        public DependencyInfo(ResourceImpl resource, boolean mandatory, boolean start, int startLevel, boolean blacklisted) {
+            this.resource = resource;
+            this.mandatory = mandatory;
+            this.start = start;
+            this.startLevel = startLevel;
+            this.blacklisted = blacklisted;
+        }
 
         @Override
         public boolean isStart() {
@@ -591,10 +772,37 @@ public class Subsystem extends ResourceImpl {
             return getUri(resource);
         }
 
+        @Override
+        public String getOriginalLocation() {
+            // resource is already overriden
+            return getUri(resource);
+        }
+
         @Override
         public boolean isDependency() {
             return !mandatory;
         }
+
+        @Override
+        public boolean isBlacklisted() {
+            return blacklisted;
+        }
+
+        @Override
+        public BundleInfo.BundleOverrideMode isOverriden() {
+            return overriden;
+        }
+
+        public void setOverriden(BundleInfo.BundleOverrideMode overriden) {
+            this.overriden = overriden;
+        }
+
+        @Override
+        public String toString() {
+            return "DependencyInfo{" +
+                    "resource=" + resource +
+                    '}';
+        }
     }
 
     Map<String, Set<String>> createPolicy(List<? extends ScopeFilter> filters) {
